@@ -1,8 +1,8 @@
 /*
   efibootmgr.c - Manipulates EFI variables as exported in /proc/efi/vars
- 
+
   Copyright (C) 2001 Dell Computer Corporation <Matt_Domsch@dell.com>
- 
+
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
@@ -27,7 +27,7 @@
     Media Device Path, File Path, length ??
     End of Hardware Device Path, length 4
     Arguments passed to elilo, as UCS-2 characters, length ??
- 
+
 */
 
 #define _GNU_SOURCE
@@ -55,10 +55,10 @@
 
 
 typedef struct _var_entry {
-	struct dirent   *name;
-	uint16_t            num;
-	efi_variable_t   var_data;
-	struct list_head list;
+	struct dirent *name;
+	uint16_t       num;
+	efi_variable_t var_data;
+	list_t         list;
 } var_entry_t;
 
 
@@ -94,9 +94,9 @@ static int
 read_boot_var_names(struct dirent ***namelist)
 {
 	int n;
-	n = scandir(PROC_DIR_EFI, namelist, select_boot_var_names, alphasort);
+	n = scandir(PROC_DIR_EFI_VARS, namelist, select_boot_var_names, alphasort);
 	if (n < 0) {
-		perror("scandir " PROC_DIR_EFI);
+		perror("scandir " PROC_DIR_EFI_VARS);
 		fprintf(stderr, "You must 'modprobe efivars' before running efibootmgr.\n");
 	}
 	return n;
@@ -107,7 +107,7 @@ static int
 read_blk_var_names(struct dirent ***namelist)
 {
 	int n;
-	n = scandir(PROC_DIR_EFI, namelist, select_blk_var_names, alphasort);
+	n = scandir(PROC_DIR_EFI_VARS, namelist, select_blk_var_names, alphasort);
 	if (n < 0)
 		perror("scandir");
 	return n;
@@ -125,7 +125,8 @@ dirent_list_length(struct dirent **namelist)
 
 static void
 read_vars(struct dirent **namelist,
-	  struct list_head *head)
+	  int num_boot_names,
+	  list_t *head)
 {
 	efi_status_t status;
 	var_entry_t *entry;
@@ -133,17 +134,19 @@ read_vars(struct dirent **namelist,
 
 	if (!namelist) return;
 
-	for (i=0; namelist[i]; i++)
+	for (i=0; i < num_boot_names; i++)
 	{
-		entry = malloc(sizeof(var_entry_t));
-		if (!entry) return;
-		memset(entry, 0, sizeof(var_entry_t));
-		
-		status = read_variable(namelist[i]->d_name, &entry->var_data);
-		if (status != EFI_SUCCESS) break;
-//		printf("Reading %s\n", namelist[i]->d_name);
-		entry->name = namelist[i];
-		list_add_tail(&entry->list, head);
+		if (namelist[i]) {
+			entry = malloc(sizeof(var_entry_t));
+			if (!entry) return;
+			memset(entry, 0, sizeof(var_entry_t));
+
+			status = read_variable(namelist[i]->d_name,
+					       &entry->var_data);
+			if (status != EFI_SUCCESS) break;
+			entry->name = namelist[i];
+			list_add_tail(&entry->list, head);
+		}
 	}
 	return;
 }
@@ -153,13 +156,15 @@ read_vars(struct dirent **namelist,
 
 
 static void
-free_dirents(struct dirent **ptr)
+free_dirents(struct dirent **ptr, int num_dirents)
 {
 	int i;
 	if (!ptr) return;
-	for (i=0; ptr[i]; i++) {
-		free(ptr[i]);
-		ptr[i] = NULL;
+	for (i=0; i < num_dirents; i++) {
+		if (ptr[i]) {
+			free(ptr[i]);
+			ptr[i] = NULL;
+		}
 	}
 	free(ptr);
 }
@@ -170,7 +175,9 @@ static int
 compare(const void *a, const void *b)
 {
 	int rc = -1;
-	uint32_t n1=*(uint32_t *)a, n2=*(uint32_t *)b;
+	uint32_t n1, n2;
+	memcpy(&n1, a, sizeof(n1));
+	memcpy(&n2, b, sizeof(n2));
 	if (n1 < n2) rc = -1;
 	if (n1 == n2) rc = 0;
 	if (n2 > n2) rc = 1;
@@ -183,11 +190,11 @@ compare(const void *a, const void *b)
   or -1 on failure.
 */
 static int
-find_free_boot_var(struct list_head *boot_list)
+find_free_boot_var(list_t *boot_list)
 {
 	int num_vars=0, i=0, found;
 	uint16_t *vars, free_number;
-	struct list_head *pos;
+	list_t *pos;
 	var_entry_t *boot;
 	list_for_each(pos, boot_list) {
 		num_vars++;
@@ -223,9 +230,9 @@ find_free_boot_var(struct list_head *boot_list)
 
 
 static void
-warn_duplicate_name(struct list_head *boot_list)
+warn_duplicate_name(list_t *boot_list)
 {
-	struct list_head *pos;
+	list_t *pos;
 	var_entry_t *boot;
 	EFI_LOAD_OPTION *load_option;
 
@@ -241,11 +248,10 @@ warn_duplicate_name(struct list_head *boot_list)
 		}
 	}
 }
-		    
 
 
 static var_entry_t *
-make_boot_var(struct list_head *boot_list)
+make_boot_var(list_t *boot_list)
 {
 	var_entry_t *boot;
 	int free_number;
@@ -263,8 +269,10 @@ make_boot_var(struct list_head *boot_list)
 	if (!boot) return NULL;
 	memset(boot, 0, sizeof(*boot));
 	boot->num = free_number;
-	make_linux_efi_variable(&boot->var_data, free_number);
-	
+	if (!make_linux_efi_variable(&boot->var_data, free_number)) {
+		free(boot);
+		return NULL;
+	}
 	write_variable(&boot->var_data);
 	list_add_tail(&boot->list, boot_list);
 	return boot;
@@ -293,6 +301,7 @@ read_boot_order(efi_variable_t *boot_order)
 		boot_order->Attributes = EFI_VARIABLE_NON_VOLATILE 
 			| EFI_VARIABLE_BOOTSERVICE_ACCESS
 			| EFI_VARIABLE_RUNTIME_ACCESS;
+		return status;
 	}
 	return EFI_SUCCESS;
 }
@@ -308,25 +317,22 @@ add_to_boot_order(uint16_t num)
 	uint64_t new_data_size;
 	uint16_t *new_data, *old_data;
 
-
 	status = read_boot_order(&boot_order);
 	if (status != EFI_SUCCESS) return status;
 
-	
 	/* We've now got an array (in boot_order.Data) of the
 	   boot order.  First add our entry, then copy the old array.
 	*/
 	old_data = (uint16_t *)&(boot_order.Data);
 	new_data_size = boot_order.DataSize + sizeof(uint16_t);
 	new_data = malloc(new_data_size);
-	
+
 	new_data[0] = num;
 	memcpy(new_data+1, old_data, boot_order.DataSize);
-	
+
 	/* Now new_data has what we need */
 	memcpy(&(boot_order.Data), new_data, new_data_size);
 	boot_order.DataSize = new_data_size;
-	
 	return write_variable(&boot_order);
 }
 
@@ -342,7 +348,7 @@ remove_from_boot_order(uint16_t num)
 
 	status = read_boot_order(&boot_order);
 	if (status != EFI_SUCCESS) return status;
-	
+
 	/* If it's empty, yea! */
 	if (!boot_order.DataSize) return EFI_SUCCESS;
 
@@ -369,7 +375,7 @@ remove_from_boot_order(uint16_t num)
 	memset(&(boot_order.Data), 0, boot_order.DataSize);
 	memcpy(&(boot_order.Data), new_data, new_data_size);
 	boot_order.DataSize = new_data_size;
-	
+
 	return write_variable(&boot_order);
 }
 
@@ -456,7 +462,7 @@ delete_boot_var(uint16_t num)
 	efi_variable_t var;
 	efi_guid_t guid = EFI_GLOBAL_VARIABLE;
 	char name[80];
-	struct list_head *pos;
+	list_t *pos, *n;
 	var_entry_t *boot;
 	sprintf(name, "Boot%04x", num);
 
@@ -468,13 +474,13 @@ delete_boot_var(uint16_t num)
 
 	if (status) return status;
 
-	list_for_each(pos, &boot_entry_list) {
+	list_for_each_safe(pos, n, &boot_entry_list) {
 		boot = list_entry(pos, var_entry_t, list);
 		if (boot->num == num) {
 			status = remove_from_boot_order(num);
 			if (status) return status;
 			list_del(&(boot->list));
-			break; /* otherwise, we loop here forever! */
+			break; /* short-circuit since it was found */
 		}
 	}
 	return EFI_SUCCESS;
@@ -482,12 +488,12 @@ delete_boot_var(uint16_t num)
 
 
 static void
-set_var_nums(const char *pattern, struct list_head *list)
+set_var_nums(const char *pattern, list_t *list)
 {
-	struct list_head *pos;
+	list_t *pos;
 	var_entry_t *var;
 	int num=0, rc;
-	
+
 	list_for_each(pos, list) {
 		var = list_entry(pos, var_entry_t, list);
 		rc = sscanf(var->name->d_name, pattern, &num);
@@ -498,9 +504,9 @@ set_var_nums(const char *pattern, struct list_head *list)
 #if 0
 static efi_variable_t *
 find_pci_scsi_disk_blk(int fd, int bus, int device, int func,
-		       struct list_head *blk_list)
+		       list_t *blk_list)
 {
-	struct list_head *pos;
+	list_t *pos;
 	int rc;
 	Scsi_Idlun idlun;
 	unsigned char host, channel, id, lun;
@@ -513,8 +519,6 @@ find_pci_scsi_disk_blk(int fd, int bus, int device, int func,
 	if (rc) return NULL;
 
 	rc = disk_get_size(fd, &size);
-	
-	
 
 	idlun_to_components(&idlun, &host, &channel, &id, &lun);
 
@@ -542,7 +546,7 @@ find_pci_scsi_disk_blk(int fd, int bus, int device, int func,
 
 
 static efi_variable_t *
-find_disk_blk(char *disk_name, struct list_head *blk_list)
+find_disk_blk(char *disk_name, list_t *blk_list)
 {
 	efi_variable_t *disk_blk = NULL;
 	int fd, rc;
@@ -551,7 +555,7 @@ find_disk_blk(char *disk_name, struct list_head *blk_list)
 	unsigned int controllernum=0, disknum=0;
 	unsigned char part=0;
 
-	fd = open(disk_name, O_RDONLY);
+	fd = open(disk_name, O_RDONLY|O_DIRECT);
 	rc = disk_get_pci(fd, &bus, &device, &func);
 	if (rc) {
 		fprintf(stderr, "disk_get_pci() failed.\n");
@@ -559,7 +563,7 @@ find_disk_blk(char *disk_name, struct list_head *blk_list)
 	}
 	rc = disk_info_from_fd(fd,
 			       &interface_type,
-			       &controllernum, 
+			       &controllernum,
 			       &disknum,
 			       &part);
 	if (rc) {
@@ -585,7 +589,7 @@ find_disk_blk(char *disk_name, struct list_head *blk_list)
 	}
 	return NULL;
 }
-#endif	
+#endif
 
 static void
 unparse_boot_order(uint16_t *order, int length)
@@ -631,7 +635,7 @@ set_boot_order()
 	efichar_from_char(var.VariableName, "BootOrder",
 			  1024);
 	memcpy(&var.VendorGuid, &guid, sizeof(guid));
-	var.Attributes = EFI_VARIABLE_NON_VOLATILE 
+	var.Attributes = EFI_VARIABLE_NON_VOLATILE
 		| EFI_VARIABLE_BOOTSERVICE_ACCESS
 		| EFI_VARIABLE_RUNTIME_ACCESS;
 
@@ -642,7 +646,7 @@ set_boot_order()
 static void
 show_boot_vars()
 {
-	struct list_head *pos;
+	list_t *pos;
 	var_entry_t *boot;
 	char description[80];
 	EFI_LOAD_OPTION *load_option;
@@ -663,9 +667,9 @@ show_boot_vars()
 			printf("* ");
 		else    printf("  ");
 		printf("%s", description);
-		
+
 		if (opts.verbose) {
-			unparse_path(text_path, path, 
+			unparse_path(text_path, path,
 				     load_option->file_path_list_length);
 			/* Print optional data */
 			optional_data_len =
@@ -679,7 +683,7 @@ show_boot_vars()
 						 load_option->file_path_list_length,
 						 optional_data_len);
 			}
-			
+
 			printf("\t%s", text_path);
 		}
 		printf("\n");
@@ -696,13 +700,17 @@ show_boot_order()
 	uint16_t *data;
 
 	status = read_boot_order(&boot_order);
-	if (status != EFI_SUCCESS) return;
-	
+
+	if (status != EFI_SUCCESS) {
+		perror("show_boot_order()");
+		return;
+	}
+
 	/* We've now got an array (in boot_order.Data) of the
 	   boot order.  First add our entry, then copy the old array.
 	*/
 	data = (uint16_t *)&(boot_order.Data);
-	if (boot_order.DataSize) 
+	if (boot_order.DataSize)
 		unparse_boot_order(data, boot_order.DataSize / sizeof(uint16_t));
 
 }
@@ -710,7 +718,7 @@ show_boot_order()
 static efi_status_t
 set_active_state()
 {
-	struct list_head *pos;
+	list_t *pos;
 	var_entry_t *boot;
 	EFI_LOAD_OPTION *load_option;
 
@@ -763,6 +771,7 @@ delete_boot_order()
 static void
 usage()
 {
+	printf("efibootmgr version %s\n", EFIBOOTMGR_VERSION);
 	printf("usage: efibootmgr [options]\n");
 	printf("\t-a | --active         sets bootnum active\n");
 	printf("\t-A | --inactive       sets bootnum inactive\n");
@@ -772,6 +781,9 @@ usage()
 	printf("\t-d | --disk disk       (defaults to /dev/sda) containing loader\n");
 	printf("\t-e | --edd [1|3|-1]   force EDD 1.0 or 3.0 creation variables, or guess\n");
 	printf("\t-E | --device num      EDD 1.0 device number (defaults to 0x80)\n");
+	printf("\t-g | --gpt            force disk with invalid PMBR to be treated as GPT\n");
+	printf("\t-H | --acpi_hid XXXX  set the ACPI HID (used with -i)\n");
+	printf("\t-i | --iface name     create a netboot entry for the named interface\n");
 	printf("\t-l | --loader name     (defaults to \\elilo.efi)\n");
 	printf("\t-L | --label label     Boot manager display label (defaults to \"Linux\")\n");
 	printf("\t-n | --bootnext XXXX   set BootNext to XXXX (hex)\n");
@@ -781,10 +793,11 @@ usage()
 	printf("\t-p | --part part        (defaults to 1) containing loader\n");
 	printf("\t-q | --quiet            be quiet\n");
 	printf("\t-t | --test filename    don't write to NVRAM, write to filename.\n");
-	printf("\t-u | --unicode | --UCS-2  pass extra args as UCS-2 (default is ASCII\n");
+	printf("\t-u | --unicode | --UCS-2  pass extra args as UCS-2 (default is ASCII)\n");
+	printf("\t-U | --acpi_uid XXXX    set the ACPI UID (used with -i)\n");
 	printf("\t-v | --verbose          print additional information\n");
 	printf("\t-V | --version          return version and exit\n");
-	printf("\t-w | --write-signature  write unique sig to disk if needed\n");
+	printf("\t-w | --write-signature  write unique sig to MBR if needed\n");
 }
 
 static void
@@ -794,11 +807,14 @@ set_default_opts()
 	opts.bootnum         = -1;   /* auto-detect */
 	opts.bootnext        = -1;   /* Don't set it */
 	opts.active          = -1;   /* Don't set it */
-	opts.edd10_devicenum = 0x80; 
+	opts.edd10_devicenum = 0x80;
 	opts.loader          = "\\elilo.efi";
 	opts.label           = "Linux";
 	opts.disk            = "/dev/sda";
+	opts.iface           = NULL;
 	opts.part            = 1;
+	opts.acpi_hid        = -1;
+	opts.acpi_uid        = -1;
 }
 
 static void
@@ -818,8 +834,11 @@ parse_opts(int argc, char **argv)
 			{"delete-bootnum",         no_argument, 0, 'B'},
 			{"create",                 no_argument, 0, 'c'},
 			{"disk",             required_argument, 0, 'd'},
+			{"iface",            required_argument, 0, 'i'},
+			{"acpi_hid",         required_argument, 0, 'H' },
 			{"edd-device",       required_argument, 0, 'E'},
 			{"edd30",            required_argument, 0, 'e'},
+			{"gpt",                    no_argument, 0, 'g'},
 			{"loader",           required_argument, 0, 'l'},
 			{"label",            required_argument, 0, 'L'},
 			{"bootnext",         required_argument, 0, 'n'},
@@ -831,14 +850,15 @@ parse_opts(int argc, char **argv)
 			{"test",             required_argument, 0, 't'},
 			{"unicode",                no_argument, 0, 'u'},
 			{"UCS-2",                  no_argument, 0, 'u'},
+			{"acpi_uid",         required_argument, 0, 'U' },
 			{"verbose",          optional_argument, 0, 'v'},
 			{"version",                no_argument, 0, 'V'},
 			{"write-signature",        no_argument, 0, 'w'},
 			{0, 0, 0, 0}
 		};
-		
+
 		c = getopt_long (argc, argv,
-				 "AaBb:cd:e:E:l:L:n:No:Op:qt:uv::Vw",
+				 "AaBb:cd:e:E:gH:i:l:L:n:No:Op:qt:uU:v::Vw",
 				 long_options, &option_index);
 		if (c == -1)
 			break;
@@ -872,6 +892,16 @@ parse_opts(int argc, char **argv)
 			rc = sscanf(optarg, "%x", &num);
 			if (rc == 1) opts.edd10_devicenum = num;
 			break;
+		case 'g':
+			opts.forcegpt = 1;
+			break;
+		case 'H':
+			rc = sscanf(optarg, "%x", &num);
+			if (rc == 1) opts.acpi_hid = num;
+			break;
+		case 'i':
+			opts.iface = optarg;
+			break;
 		case 'l':
 			opts.loader = optarg;
 			break;
@@ -905,6 +935,10 @@ parse_opts(int argc, char **argv)
 			opts.unicode = 1;
 			break;
 
+		case 'U':
+			rc = sscanf(optarg, "%x", &num);
+			if (rc == 1) opts.acpi_uid = num;
+			break;
 		case 'v':
 			opts.verbose = 1;
 			if (optarg) {
@@ -936,18 +970,13 @@ parse_opts(int argc, char **argv)
 }
 
 
-int 
+int
 main(int argc, char **argv)
 {
 	struct dirent  **boot_names = NULL;
 	var_entry_t *new_boot = NULL;
-	int num;
+	int num, num_boot_names=0;
 
-	if (getuid() != 0) {
-		printf("efibootmgr must be run as root.\n");
-		exit(1);
-	}
-	
 	set_default_opts();
 	parse_opts(argc, argv);
 	if (opts.showversion) {
@@ -955,35 +984,42 @@ main(int argc, char **argv)
 		return 0;
 	}
 
-	
+	if (opts.iface && opts.acpi_hid == -1 && opts.acpi_uid == -1) {
+		fprintf(stderr, "\nYou must specify the ACPI HID and UID when using -i.\n\n");
+		return 1;
+	}
+
 	if (!opts.testfile) {
-		read_boot_var_names(&boot_names);
-		read_vars(boot_names, &boot_entry_list);
+		num_boot_names = read_boot_var_names(&boot_names);
+		read_vars(boot_names, num_boot_names, &boot_entry_list);
 		set_var_nums("Boot%04x-%*s", &boot_entry_list);
-		
+
 		if (opts.delete_boot) {
-			delete_boot_var(opts.bootnum);
+			if (opts.bootnum == -1)
+				fprintf(stderr, "\nYou must specify a boot entry to delete (see the -b option).\n\n");
+			else
+				delete_boot_var(opts.bootnum);
 		}
-		
+
 		if (opts.active >= 0) {
 			set_active_state();
 		}
 	}
-		
+
 	if (opts.create) {
 		warn_duplicate_name(&boot_entry_list);
 		new_boot = make_boot_var(&boot_entry_list);
 		/* Put this boot var in the right BootOrder */
-		if (!opts.testfile) 
+		if (!opts.testfile && new_boot)
 			add_to_boot_order(new_boot->num);
 	}
 
 	if (!opts.testfile) {
-		
+
 		if (opts.delete_bootorder) {
 			delete_boot_order();
 		}
-		
+
 		if (opts.bootorder) {
 			set_boot_order();
 		}
@@ -1010,7 +1046,7 @@ main(int argc, char **argv)
 			show_boot_vars();
 		}
 	}
-	free_dirents(boot_names);
+	free_dirents(boot_names, num_boot_names);
 	return 0;
-} 
+}
 
