@@ -1,8 +1,8 @@
 /*
   efi.[ch] - Manipulates EFI variables as exported in /proc/efi/vars
- 
-  Copyright (C) 2001 Dell Computer Corporation <Matt_Domsch@dell.com>
- 
+
+  Copyright (C) 2001,2003 Dell Computer Corporation <Matt_Domsch@dell.com>
+
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
@@ -29,6 +29,19 @@
 #include <fcntl.h>
 #include <limits.h>
 #include <unistd.h>
+#include <dirent.h>
+#include <sys/socket.h>
+#include <sys/types.h>
+#include <sys/ioctl.h>
+#include <linux/sockios.h>
+#include <net/if.h>
+
+typedef unsigned long long u64; /* hack, so we may include kernel's ethtool.h */
+typedef __uint32_t u32;         /* ditto */
+typedef __uint16_t u16;         /* ditto */
+typedef __uint8_t u8;           /* ditto */
+
+#include <linux/ethtool.h>
 #include "efi.h"
 #include "efichar.h"
 #include "scsi_ioctls.h"
@@ -45,13 +58,15 @@ load_option_path(EFI_LOAD_OPTION *option)
 		 + efichar_strsize(option->description)); /* Description */
 }
 
-void
+char *
 efi_guid_unparse(efi_guid_t *guid, char *out)
 {
-        sprintf(out, "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
-                guid->data1, guid->data2, guid->data3,
-                guid->data4[0], guid->data4[1], guid->data4[2], guid->data4[3],
-                guid->data4[4], guid->data4[5], guid->data4[6], guid->data4[7]);
+	sprintf(out, "%02x%02x%02x%02x-%02x%02x-%02x%02x-%02x%02x-%02x%02x%02x%02x%02x%02x",
+		guid->b[3], guid->b[2], guid->b[1], guid->b[0],
+		guid->b[5], guid->b[4], guid->b[7], guid->b[6],
+		guid->b[8], guid->b[9], guid->b[10], guid->b[11],
+		guid->b[12], guid->b[13], guid->b[14], guid->b[15]);
+        return out;
 }
 
 
@@ -63,11 +78,11 @@ read_variable(char *name, efi_variable_t *var)
 	int fd;
 	size_t readsize;
 	if (!name || !var) return EFI_INVALID_PARAMETER;
-	
-	newnamesize = strlen(PROC_DIR_EFI) + strlen(name) + 1;
+
+	newnamesize = strlen(PROC_DIR_EFI_VARS) + strlen(name) + 1;
 	newname = malloc(newnamesize);
 	if (!newname) return EFI_OUT_OF_RESOURCES;
-	sprintf(newname, "%s%s", PROC_DIR_EFI,name);
+	sprintf(newname, "%s%s", PROC_DIR_EFI_VARS,name);
 	fd = open(newname, O_RDONLY);
 	if (fd == -1) {
 		free(newname);
@@ -78,7 +93,6 @@ read_variable(char *name, efi_variable_t *var)
 		free(newname);
 		close(fd);
 		return EFI_INVALID_PARAMETER;
-		
 	}
 	close(fd);
 	free(newname);
@@ -106,7 +120,72 @@ write_variable_to_file(efi_variable_t *var)
 	close(fd);
 	return EFI_SUCCESS;
 }
+/**
+ * select_variable_names()
+ * @d - dirent to compare against
+ *
+ * This ignores "." and ".." entries, and selects all others.
+ */
 
+static int
+select_variable_names(const struct dirent *d)
+{
+	if (!strcmp(d->d_name, ".") ||
+	    !strcmp(d->d_name, ".."))
+		return 0;
+	return 1;
+}
+
+/**
+ * find_write_victim()
+ * @var - variable to be written
+ * @file - name of file to open for writing @var is returned.
+ *
+ * This ignores "." and ".." entries, and selects all others.
+ */
+static char *
+find_write_victim(efi_variable_t *var, char file[PATH_MAX])
+{
+	struct dirent **namelist = NULL;
+	int i, n, found=0;
+	char testname[PATH_MAX], *p;
+
+	memset(testname, 0, sizeof(testname));
+	n = scandir(PROC_DIR_EFI_VARS, &namelist,
+		    select_variable_names, alphasort);
+	if (n < 0) {
+		perror("scandir " PROC_DIR_EFI_VARS);
+		fprintf(stderr, "You must 'modprobe efivars' first.\n");
+		return NULL;
+	}
+
+	p = testname;
+	efichar_to_char(p, var->VariableName, PATH_MAX);
+	p += strlen(p);
+	p += sprintf(p, "-");
+	efi_guid_unparse(&var->VendorGuid, p);
+
+	for (i=0; i<n; i++) {
+		if (namelist[i] &&
+		    strncmp(testname, namelist[i]->d_name, sizeof(testname))) {
+			found++;
+			sprintf(file, "%s%s", PROC_DIR_EFI_VARS,
+				namelist[i]->d_name);
+			break;
+		}
+	}
+
+	while (n--) {
+		if (namelist[n]) {
+			free(namelist[n]);
+			namelist[n] = NULL;
+		}
+	}
+	free(namelist);
+
+	if (!found) return NULL;
+	return file;
+}
 
 
 efi_status_t
@@ -114,12 +193,16 @@ write_variable(efi_variable_t *var)
 {
 	int fd;
 	size_t writesize;
-	char buffer[PATH_MAX];
-	char name[]="/proc/efi/vars/Efi-47c7b226-c42a-11d2-8e57-00a0c969723b";
-	if (!var) return EFI_INVALID_PARAMETER;
+	char buffer[PATH_MAX], name[PATH_MAX], *p = NULL;
 
+	if (!var) return EFI_INVALID_PARAMETER;
 	if (opts.testfile) return write_variable_to_file(var);
-	
+	memset(buffer, 0, sizeof(buffer));
+	memset(name, 0, sizeof(name));
+
+	p = find_write_victim(var, name);
+	if (!p) return EFI_INVALID_PARAMETER;
+
 	fd = open(name, O_WRONLY);
 	if (fd == -1) {
 		sprintf(buffer, "write_variable():open(%s)", name);
@@ -135,7 +218,7 @@ write_variable(efi_variable_t *var)
 #endif
 		close(fd);
 		return EFI_INVALID_PARAMETER;
-		
+
 	}
 	close(fd);
 	return EFI_SUCCESS;
@@ -173,7 +256,7 @@ get_edd_version()
 	memset(&var, 0, sizeof(efi_variable_t));
 	efi_guid_unparse(&guid, text_guid);
 	sprintf(name, "blk0-%s", text_guid);
-	
+
 	status = read_variable(name, &var);
 	if (status != EFI_SUCCESS) {
 		return 0;
@@ -183,105 +266,138 @@ get_edd_version()
 	return rc;
 }
 
-
-
-
 /*
   EFI_DEVICE_PATH, 0x01 (Hardware), 0x04 (Vendor), length 0x0018
   This needs to know what EFI device has the boot device.
 */
 static uint16_t
-make_edd10_device_path(void *buffer, uint32_t hardware_device)
+make_edd10_device_path(void *dest, uint32_t hardware_device)
 {
-	VENDOR_DEVICE_PATH *hw = buffer;
+	VENDOR_DEVICE_PATH *hw;
+	char buffer[EDD10_HARDWARE_VENDOR_PATH_LENGTH];
 	efi_guid_t guid = EDD10_HARDWARE_VENDOR_PATH_GUID;
-	uint32_t *data = (uint32_t *)hw->data;
+	uint32_t *data;
+	memset(buffer, 0, sizeof(buffer));
+	hw = (VENDOR_DEVICE_PATH *)buffer;
+	data = (uint32_t *)hw->data;
 	hw->type = 0x01; /* Hardware Device Path */
 	hw->subtype = 0x04; /* Vendor */
-	hw->length = 24;
+	hw->length = EDD10_HARDWARE_VENDOR_PATH_LENGTH;
 	memcpy(&(hw->vendor_guid), &guid, sizeof(guid));
 	*data = hardware_device;
+	memcpy(dest, buffer, hw->length);
 	return hw->length;
 }
 
-
-
 static uint16_t
-make_end_device_path(void *buffer)
+make_end_device_path(void *dest)
 {
-	END_DEVICE_PATH *p = buffer;
-	p->type = 0x7F; /* End of Hardware Device Path */
-	p->subtype = 0xFF; /* End Entire Device Path */
-	p->length = sizeof(*p);
-	return p->length;
-}
-
-
-static uint16_t
-make_acpi_device_path(void *buffer, uint32_t _HID, uint32_t _UID)
-{
-	ACPI_DEVICE_PATH *p = buffer;
-	p->type = 2;
-	p->subtype = 1;
-	p->length = sizeof(*p);
-	p->_HID = _HID;
-	p->_UID = _UID;
-	return p->length;
+	END_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 0x7F; /* End of Hardware Device Path */
+	p.subtype = 0xFF; /* End Entire Device Path */
+	p.length = sizeof(p);
+	memcpy(dest, &p, p.length);
+	return p.length;
 }
 
 static uint16_t
-make_pci_device_path(void *buffer, uint8_t device, uint8_t function)
+make_acpi_device_path(void *dest, uint32_t _HID, uint32_t _UID)
 {
-	PCI_DEVICE_PATH *p = buffer;
-	p->type = 1;
-	p->subtype = 1;
-	p->length   = sizeof(*p);
-	p->device   = device;
-	p->function = function;
-	return p->length;
+	ACPI_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 2;
+	p.subtype = 1;
+	p.length = sizeof(p);
+	p._HID = _HID;
+	p._UID = _UID;
+	memcpy(dest, &p, p.length);
+	return p.length;
 }
 
 static uint16_t
-make_scsi_device_path(void *buffer, uint16_t id, uint16_t lun)
+make_mac_addr_device_path(void *dest, char *mac, uint8_t iftype)
 {
-	SCSI_DEVICE_PATH *p = buffer;
-	p->type = 3;
-	p->subtype = 2;
-	p->length   = sizeof(*p);
-	p->id       = id;
-	p->lun      = lun;
-	return p->length;
+
+        int i;
+	MAC_ADDR_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 3;
+	p.subtype = 11;
+	p.length = sizeof(p);
+	for (i=0; i < 14; i++) {
+		p.macaddr[i] = mac[i];
+	}
+	p.iftype = iftype;
+	memcpy(dest, &p, p.length);
+	return p.length;
 }
 
 static uint16_t
-make_harddrive_device_path(void *buffer, uint32_t num, uint64_t start, uint64_t size,
+make_pci_device_path(void *dest, uint8_t device, uint8_t function)
+{
+	PCI_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 1;
+	p.subtype = 1;
+	p.length   = sizeof(p);
+	p.device   = device;
+	p.function = function;
+	memcpy(dest, &p, p.length);
+	return p.length;
+}
+
+static uint16_t
+make_scsi_device_path(void *dest, uint16_t id, uint16_t lun)
+{
+	SCSI_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 3;
+	p.subtype = 2;
+	p.length   = sizeof(p);
+	p.id       = id;
+	p.lun      = lun;
+	memcpy(dest, &p, p.length);
+	return p.length;
+}
+
+static uint16_t
+make_harddrive_device_path(void *dest, uint32_t num, uint64_t start, uint64_t size,
 			   uint8_t *signature,
 			   uint8_t mbr_type, uint8_t signature_type)
 {
-	HARDDRIVE_DEVICE_PATH *p = buffer;
-	p->type = 4;
-	p->subtype = 1;
-	p->length   = sizeof(*p);
-	p->part_num = num;
-	p->start = start;
-	p->size = size;
-	if (signature) memcpy(p->signature, signature, 16);
-	p->mbr_type = mbr_type;
-	p->signature_type = signature_type;
-	return p->length;
+	HARDDRIVE_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 4;
+	p.subtype = 1;
+	p.length   = sizeof(p);
+	p.part_num = num;
+	p.start = start;
+	p.size = size;
+	if (signature) memcpy(p.signature, signature, 16);
+	p.mbr_type = mbr_type;
+	p.signature_type = signature_type;
+	memcpy(dest, &p, p.length);
+	return p.length;
 }
 
 static uint16_t
-make_file_path_device_path(void *buffer, efi_char16_t *name)
+make_file_path_device_path(void *dest, efi_char16_t *name)
 {
-	FILE_PATH_DEVICE_PATH *p = buffer;
+	FILE_PATH_DEVICE_PATH *p;
+	char buffer[1024];
 	int namelen  = efichar_strlen(name, -1);
 	int namesize = efichar_strsize(name);
+
+	memset(buffer, 0, sizeof(buffer));
+	p = (FILE_PATH_DEVICE_PATH *)buffer;
 	p->type      = 4;
 	p->subtype   = 4;
 	p->length    = 4 + namesize;
 	efichar_strncpy(p->path_name,
 			name, namelen);
+
+	memcpy(dest, buffer, p->length);
 	return p->length;
 
 }
@@ -306,31 +422,134 @@ make_edd30_device_path(int fd, void *buffer)
 	if (rc) return 0;
 	idlun_to_components(&idlun, &host, &channel, &id, &lun);
 
-	
 	p += make_acpi_device_path      (p, EISAID_PNP0A03, bus);
 	p += make_pci_device_path       (p, device, function);
 	p += make_scsi_device_path      (p, id, lun);
 	return ((void *)p - buffer);
 }
 
+/**
+ * make_disk_load_option()
+ * @disk disk
+ *
+ * Returns 0 on error, length of load option created on success.
+ */
+char *make_disk_load_option(char *p, char *disk)
+{
+    int disk_fd=0;
+    char buffer[80];
+    char signature[16];
+    int rc, edd_version=0;
+    uint8_t mbr_type=0, signature_type=0;
+    uint64_t start=0, size=0;
+    efi_char16_t os_loader_path[40];
+
+    memset(signature, 0, sizeof(signature));
+
+    disk_fd = open(opts.disk, O_RDWR);
+    if (disk_fd == -1) {
+        sprintf(buffer, "Could not open disk %s", opts.disk);
+	perror(buffer);
+	return 0;
+    }
+
+    if (opts.edd_version) {
+        edd_version = get_edd_version();
+
+	if (edd_version == 3) {
+	    p += make_edd30_device_path(disk_fd, p);
+	}
+	else if (edd_version == 1) {
+	    p += make_edd10_device_path(p, opts.edd10_devicenum);
+	}
+    }
+
+    rc = disk_get_partition_info (disk_fd, opts.part,
+				  &start, &size, signature,
+				  &mbr_type, &signature_type);
+
+    close(disk_fd);
+
+    if (rc) {
+        fprintf(stderr, "Error: no partition information on disk %s.\n"
+		"       Cowardly refusing to create a boot option.\n",
+		opts.disk);
+	return 0;
+    }
+
+    p += make_harddrive_device_path (p, opts.part,
+				     start, size,
+				     signature,
+				     mbr_type, signature_type);
+
+    efichar_from_char(os_loader_path, opts.loader, sizeof(os_loader_path));
+    p += make_file_path_device_path (p, os_loader_path);
+    p += make_end_device_path       (p);
+
+    return(p);
+}
+
+/**
+ * make_net_load_option()
+ * @data - load option returned
+ *
+ * Returns 0 on error, length of load option created on success.
+ */
+char *make_net_load_option(char *p, char *iface)
+{
+    /* copied pretty much verbatim from the ethtool source */
+    int fd = 0, err;
+    int bus, slot, func;
+    struct ifreq ifr;
+    struct ethtool_drvinfo drvinfo;
+
+    memset(&ifr, 0, sizeof(ifr));
+    strcpy(ifr.ifr_name, iface);
+    drvinfo.cmd = ETHTOOL_GDRVINFO;
+    ifr.ifr_data = (caddr_t)&drvinfo;
+    /* Open control socket */
+    fd = socket(AF_INET, SOCK_DGRAM, 0);
+    if (fd < 0) {
+        perror("Cannot get control socket");
+    }
+    err = ioctl(fd, SIOCETHTOOL, &ifr);
+    if (err < 0) {
+        perror("Cannot get driver information");
+    }
+
+    err = sscanf(drvinfo.bus_info, "%2x:%2x.%x", &bus, &slot, &func);
+    if (err == 0) {
+        perror("Couldn't parse device location string.");
+    }
+
+    p += make_acpi_device_path(p, opts.acpi_hid, opts.acpi_uid);
+    p += make_pci_device_path(p, (uint8_t)slot, (uint8_t)func);
+
+    err = ioctl(fd, SIOCGIFHWADDR, &ifr);
+    if (err < 0) {
+        perror("Cannot get hardware address.");
+    }
+
+    p += make_mac_addr_device_path(p, ifr.ifr_ifru.ifru_hwaddr.sa_data, 0);
+    p += make_end_device_path       (p);
+
+    return(p);
+}
+
+/**
+ * make_linux_load_option()
+ * @data - load option returned
+ *
+ * Returns 0 on error, length of load option created on success.
+ */
 static unsigned long
 make_linux_load_option(void *data)
 {
-
 	EFI_LOAD_OPTION *load_option = data;
-	char buffer[80];
-	int disk_fd=0;
 	char *p = data, *q;
-	efi_char16_t description[40];
-	efi_char16_t os_loader_path[40];
-	int rc, edd_version=0;
-	uint64_t start=0, size=0;
-	uint8_t mbr_type=0, signature_type=0;
-	char signature[16];
+	efi_char16_t description[64];
 	unsigned long datasize=0;
 
-	memset(signature, 0, sizeof(signature));
-
 	/* Write Attributes */
 	if (opts.active) load_option->attributes = LOAD_OPTION_ACTIVE;
 	else             load_option->attributes = 0;
@@ -342,47 +561,18 @@ make_linux_load_option(void *data)
 	memset(description, 0, sizeof(description));
 	efichar_from_char(description, opts.label, sizeof(description));
 	efichar_strncpy(load_option->description, description, sizeof(description));
-	p += efichar_strsize(load_option->description); 
+	p += efichar_strsize(load_option->description);
 
 	q = p;
 
-
-
-	disk_fd = open(opts.disk, O_RDWR);
-	if (disk_fd == -1) {
-		sprintf(buffer, "Could not open disk %s", opts.disk);
-		perror(buffer);
-		return 0;
+	if (opts.iface) {
+	      p = (char *)make_net_load_option(p, opts.iface);
 	}
 
-	if (opts.edd_version) {
-		edd_version = get_edd_version();
-
-		if (edd_version == 3) {
-			p += make_edd30_device_path(disk_fd, p);
-		}
-		else if (edd_version == 1) {
-			p += make_edd10_device_path(p, opts.edd10_devicenum);
-		}
+	else {
+	      p = (char *)make_disk_load_option(p, opts.iface);
 	}
-		
-	rc = disk_get_partition_info (disk_fd, opts.part,
-				      &start, &size, signature,
-				      &mbr_type, &signature_type);
-	
-	close(disk_fd);
-	
-	if (rc) return 0;
 
- 	p += make_harddrive_device_path (p, opts.part,
-					 start, size,
-					 signature,
-					 mbr_type, signature_type);
-
-	efichar_from_char(os_loader_path, opts.loader, sizeof(os_loader_path));
-	p += make_file_path_device_path (p, os_loader_path);
-	p += make_end_device_path       (p);
-	
 	load_option->file_path_list_length = p - q;
 
 	datasize = (uint8_t *)p - (uint8_t *)data;
@@ -409,11 +599,11 @@ append_extra_args_ascii(void *data, unsigned long maxchars)
 		p += strlen(p);
 		appended=1;
 
-		usedchars = p - (char *)data; 
+		usedchars = p - (char *)data;
 
 		/* Put a space between args */
 		if (i < (opts.argc-1)) {
-			
+
 			p = strncpy(p, " ", maxchars-usedchars-1);
 			p += strlen(p);
 			usedchars = p - (char *)data;
@@ -448,7 +638,7 @@ append_extra_args_unicode(void *data, unsigned long maxchars)
 				sizeof(efi_char16_t);
 		}
 	}
-	
+
 	if (appended) return efichar_strsize( (efi_char16_t *)data );
 	return 0;
 }
@@ -465,7 +655,7 @@ append_extra_args(void *data, unsigned long maxchars)
 
 
 
-void
+int
 make_linux_efi_variable(efi_variable_t *var,
 			unsigned int free_number)
 {
@@ -475,10 +665,10 @@ make_linux_efi_variable(efi_variable_t *var,
 	unsigned long load_option_size = 0, opt_data_size=0;
 
 	memset(buffer,    0, sizeof(buffer));
-	
+
 	/* VariableName needs to be BootXXXX */
 	sprintf(buffer, "Boot%04x", free_number);
-	
+
 	efichar_from_char(var->VariableName, buffer, 1024);
 
 	memcpy(&(var->VendorGuid), &guid, sizeof(guid));
@@ -491,10 +681,12 @@ make_linux_efi_variable(efi_variable_t *var,
 
 	load_option_size =  make_linux_load_option(var->Data);
 
+	if (!load_option_size) return 0;
+
 	/* Set OptionalData (passed as binary to the called app) */
 	optional_data = var->Data + load_option_size;
 	opt_data_size = append_extra_args(optional_data,
 				  sizeof(var->Data) - load_option_size);
 	var->DataSize = load_option_size + opt_data_size;
-	return;
+	return var->DataSize;
 }
